/*
 *  Created on: 14.07.2009
 *      Author: afestini
 *
 *  Binary streams act mostly like regular streams, but are meant to be used for binary
 *  data instead of text.
 *
 *  Trying to read from an empty stream is not an error, but will set a fail flag and
 *  leave the destination buffer/variable untouched (or fill with a partial read).
 *
 *  Bytes can be skipped by calling discard(size).
 *
 *  InStreams must be initialized with existing data when constructed (at this point
 *  by a vector which will be copied to the internal buffer, a more efficient option
 *  would be nice).
 */

#ifndef BINARY_STREAM_H_
#define BINARY_STREAM_H_

#include <vector>
#include <string>

class BinaryStream
{
public:
	virtual ~BinaryStream() {}

	typedef char DataType;
	typedef std::vector<DataType> Buffer;

	const DataType* begin() const { return _readPtr; }
	const DataType* end() const { return _endPtr; }
	size_t size() const { return _endPtr - _readPtr; }
	bool fail() const { return _fail; }
	void clear() { _fail = false; }
	bool eof() const { return _readPtr >= _endPtr; }

protected:
	BinaryStream() : _readPtr(0), _endPtr(0), _fail(false) {}
	BinaryStream(const DataType* begin, const DataType* end)
		: _readPtr(begin), _endPtr(end), _fail(false) {}

	const DataType *_readPtr, *_endPtr;
	bool _fail;
	Buffer _buffer;
};

class BinaryInStream : public BinaryStream
{
public:
	BinaryInStream(const Buffer& buf) : BinaryStream(&*buf.begin(), &*buf.end()) {}
	BinaryInStream(const void* begin, const void* end) : BinaryStream((DataType*)begin, (DataType*)end) {}
	BinaryInStream(const void* data, size_t size) : BinaryStream((DataType*)data, (DataType*)data+size) {}

	void copyDataToInternalBuffer() { _buffer.assign(_readPtr, _endPtr); }

	BinaryInStream& read(void* dst, size_t size)
	{
		size_t rest = end() - begin();
		if ( rest < size )
		{
			size = rest;
			_fail = true;
		}

		if (size > 0)
		{
			memcpy(dst, _readPtr, size);
			_readPtr += size;
		}
		return *this;
	}

	BinaryInStream& discard(size_t bytes)
	{
		if (bytes > size())
		{
			bytes = size();
			_fail = true;
		}
		_readPtr += bytes;
		return *this;
	}

	BinaryInStream& operator>>(std::string& dst)
	{
		dst = std::string(begin());
		return discard(dst.length()+1);
	}

	BinaryInStream& getString(std::string& dst, const std::string& delims = "\0", size_t maxChars = 0)
	{
		const DataType* searchEnd = maxChars ? begin()+maxChars : end();
		const DataType* pos = std::find_first_of(begin(), searchEnd, delims.begin(), delims.end());

		dst = std::string(begin(), pos);
		return discard( dst.length() + (pos == searchEnd ? 0 : 1) );
	}
};

//Back to template, since the compiler gets confused by typedefs and defaults
//to int version for uint16, etc. (so keep in mind _WHAT_ you put here)
template<typename T>
inline BinaryInStream& operator>>(BinaryInStream& strm, T& dst) { return strm.read(&dst, sizeof(T)); }

template<typename T, int N>
inline BinaryInStream& operator>>(BinaryInStream& strm, T (&dst)[N])
{
	for (size_t i=0; i<N; ++i)
		strm >> dst[i];
	return strm;
}

template<typename T>
inline BinaryInStream& operator&(BinaryInStream& strm, T& dst) { return strm >> dst; }


class BinaryOutStream : public BinaryStream
{
public:
	BinaryOutStream( size_t initialSize = 0 ) {
		reserve(initialSize); _readPtr = _endPtr = &_buffer[0]; }

	void reserve( size_t size ) { _buffer.reserve(size); }

	BinaryOutStream& write(const void* src, size_t size)
	{
		size_t pos = _buffer.size();
		_buffer.resize( pos + size );
		_readPtr = &_buffer[0];
		_endPtr = &*_buffer.end();
		memcpy(&_buffer[pos], src, size);
		return *this;
	}

	inline BinaryOutStream& operator<<(const std::string& src) {
		return write(src.c_str(), src.length()+1);
	}
};

template<typename T>
inline BinaryOutStream& operator<<(BinaryOutStream& strm, T src)
{
	return strm.write(&src, sizeof(T));
}

template<typename T, int N>
inline BinaryOutStream& operator<<(BinaryOutStream& strm, T const(&src)[N])
{
	for (size_t i=0; i<N; ++i)
		strm << src[i];
	return strm;
}

template<>
inline BinaryOutStream& operator<<(BinaryOutStream& strm, char* src)
{
	return strm.write(src, strlen(src)+1);
}

template<>
inline BinaryOutStream& operator<<(BinaryOutStream& strm, const char* src)
{
	return strm.write(src, strlen(src)+1);
}

template<typename T>
inline BinaryOutStream& operator&(BinaryOutStream& strm, T& dst) { return strm << dst; }


template<typename T>
BinaryStream& operator&(BinaryStream& s, T& dst)
{
	BinaryInStream* is = dynamic_cast<BinaryInStream*>(&s);
	BinaryOutStream* os = dynamic_cast<BinaryOutStream*>(&s);
	if (is) (*is) & dst;
	if (os) (*os) & dst;
	return s;
}

#endif /* BINARY_STREAM_H_ */
