/*
 * menu.h
 *
 *  Created on: 06.08.2009
 *      Author: afestini
 *
 *  Quick and dirty system for console based user interface.
 *
 *  void addEntry(const std::string& txt, T callback)
 *  	Add menu entry and function/functor to call on selection
 *
 *  void clear()
 *  	Clear all menu entries

	void clearContent()
		Clear content area

	void addLine(const std::string& msg)
		Add line of text to content area (will "scroll" if full)

	void setMessage(const std::string& msg)
		Message to be displayed in message line (for feedback messages)

	void display()
		Refresh display of content and menu

	void selection()
		Wait for user selection (use select on stdin first if necessary)
 */

#ifndef MENU_H_
#define MENU_H_

#include "simplecallback.h"
#include <iostream>
#include <iterator>
#include <vector>
#include <list>

class ConsoleMenu
{
	class MenuEntry
	{
	public:
		template<class T>
		MenuEntry(const std::string& txt, T& cb)
			: text(txt), callback(cb) {}

		std::string text;
		Callback callback;
	};

	static const size_t contentSize = 14;
	std::vector<MenuEntry> entries;
	std::list<std::string> content;
	std::string message;

public:
	template<class T>
	void addEntry(const std::string& txt, T callback) {
		entries.push_back( MenuEntry(txt, callback) );
	}

	void clear() { entries.clear(); }

	void clearContent() { content.clear(); }

	void addLine(const std::string& msg)
	{
		content.push_back(msg);
		if (content.size() > contentSize)
			content.pop_front();
	}

	void setMessage(const std::string& msg) { message = msg; }

	void display()
	{
		system("clear");

		std::cout << std::string(80, '*') << '\n';
		for (std::list<std::string>::iterator it = content.begin(); it != content.end(); ++it )
		{
			if (it->size() > 76) it->resize(76);
			std::cout << "* " << *it << std::string(76 - it->size(), ' ') << " *\n";
		}
		for (size_t i = content.size(); i<contentSize; ++i)
			std::cout << "* " << std::string(76, ' ') << " *\n";

		std::cout << std::string(80, '*') << '\n';
		std::cout << message << "\n\n";

		unsigned numLines = entries.size()/3 + (entries.size() % 3 > 0 ? 1 : 0);

		for (size_t i = 0; i < numLines; ++i)
		{
			std::cout << "    ";
			for ( size_t r = 0; r < 3; ++r)
			{
				size_t idx = i + r*numLines;
				if (idx < entries.size() )
				{
					int numWidth = i > 9 ? 4 : 3;
					std::string opt = entries[idx].text.substr(0, 23-numWidth);
					std::cout << idx << ") " << opt;
					if (r<2) std::cout << std::string(25 - numWidth - opt.length(),' ');
				}
			}
			std::cout << " \n";
		}
		(std::cout << "\nSelection: ").flush();
	}

	void selection()
	{
		unsigned selection = 0;
		std::cin >> selection;

		if (!std::cin.fail() && selection < entries.size())
		{
			message.clear();
			entries[selection].callback();
		}
		else
			setMessage("Invalid selection!");

		std::cin.sync();
	}
};

#endif /* MENU_H_ */
