/*
 * eventhandling.h
 *
 *  Created on: 05.08.2009
 *      Author: afestini
 *
 *  Simple event handling implementation.
 *
 *  Objects passed as events are derived from Event
 *  Classes able to handle events are derived from EventHandler
 *
 *  Handler functions for events need to be void func(const EventType&)
 *  and can be member or regular functions.
 *
 *  Handler functions are registered by calling registerEventFunc(FunctionPtr)
 *  and can be unregistered by calling unregisterEventFunc<EventType>();
 *
 *  handlerObj.handleEvent(eventObject) will dispatch the event to the
 *  function registered for this object type (automatically determined by it's parameter)
 */


#ifndef EVENT_HANDLING_INCLUDED
#define EVENT_HANDLING_INCLUDED

#include <string>
#include <map>
#include <typeinfo>

class TypeInfo
{
public:
	explicit TypeInfo(const std::type_info& info) : _typeInfo(info) {}

	bool operator < (const TypeInfo& rhs) const	{
		return _typeInfo.before(rhs._typeInfo) != 0;
	}

private:
	const std::type_info& _typeInfo;
};

class Event
{
protected:
	Event() {}
	virtual ~Event() {}
};

class HandlerFunctionBase
{
public:
	virtual ~HandlerFunctionBase() {}
	void exec(const Event& event) {call(event);}

private:
	virtual void call(const Event&) = 0;
};

template < class T, class EventT >
class MemberFunctionHandler : public HandlerFunctionBase
{
public:
	typedef void (T::*MemberFunc)(const EventT&);
	MemberFunctionHandler(T* instance, MemberFunc memFn) : _instance(instance), _function(memFn) {}

	void call(const Event& event) { (_instance->*_function)(static_cast< const EventT& >(event)); }

private:
  T* _instance;
  MemberFunc _function;
};

template < class EventT >
class StaticFunctionHandler : public HandlerFunctionBase
{
public:
	typedef void (*Func)(const EventT&);
	StaticFunctionHandler(Func Fn) : _function(Fn) {}

	void call(const Event& event) { _function(static_cast< const EventT& >(event)); }

private:
  Func _function;
};

class EventHandler
{
public:
	~EventHandler()
	{
		for ( Handlers::iterator it = _handlers.begin(); it != _handlers.end(); ++it )
			delete it->second;
		_handlers.clear();
	}

	void handleEvent(const Event& event)
	{
		Handlers::iterator it = _handlers.find(TypeInfo(typeid(event)));
		if( it != _handlers.end() )
			it->second->exec(event);
	}

	template < class D, class B, class EventT >
	void registerEventFunc(D* obj, void(B::*memFn)(const EventT&))
	{
		unregisterEventFunc<EventT>();
		_handlers[TypeInfo(typeid(EventT))] = new MemberFunctionHandler< B, EventT >(obj, memFn);
	}

	template < class EventT >
	void registerEventFunc(void(*Fn)(const EventT&))
	{
		unregisterEventFunc<EventT>();
		_handlers[TypeInfo(typeid(EventT))] = new StaticFunctionHandler< EventT >(Fn);
	}

	template <class EventT>
	void unregisterEventFunc()
	{
		Handlers::iterator it = _handlers.find(TypeInfo(typeid(EventT)));
		if( it != _handlers.end() )
		{
			delete it->second;
			_handlers.erase(it);
		}
	}

private:
	typedef std::map< TypeInfo, HandlerFunctionBase* > Handlers;
	Handlers _handlers;
};

#endif
