/*
 * interthreadcomm.h
 *
 *  Created on: 04.08.2009
 *      Author: afestini
 *
 *  Only implements QueueWithPipeNotify at this point. A channel needs to implement
 *  send and receive functions. If the channel should be usable for "select", it must
 *  return the file handle through getFD.
 */

#ifndef INTERTHREADCOMM_H_
#define INTERTHREADCOMM_H_

#include <queue>
#include "scopedlock.h"
#include <memory>

template <class Message>
class InterThreadChannel
{
public:
	virtual ~InterThreadChannel() {}

	virtual bool send(Message*) = 0;
	virtual std::auto_ptr<Message> receive() = 0;
};

class SelectableChannel
{
public:
	virtual ~SelectableChannel() {}

	virtual int getFD() const = 0;
};

template <class Message>
class QueueWithPipeNotify : public InterThreadChannel<Message>, SelectableChannel
{
	int pipe_[2];
	pthread_mutex_t mutex_;
	std::queue<Message*> msgQueue_;

public:
	QueueWithPipeNotify()
	{
		pthread_mutex_init(&mutex_, 0);
		pipe(pipe_);
	}

	~QueueWithPipeNotify()
	{
		ScopedLock l(mutex_);
		while (!msgQueue_.empty())
		{
			delete msgQueue_.front();
			msgQueue_.pop();
		}
		close(pipe_[0]);
		close(pipe_[1]);
		l.unlock();

		pthread_mutex_destroy(&mutex_);
	}

	virtual int getFD() const { return pipe_[0]; }

	virtual bool send(Message* msg)
	{
		ScopedLock l(mutex_);
		msgQueue_.push(msg);
		write( pipe_[1], "", 1);
		return true;
	}

	virtual std::auto_ptr<Message> receive()
	{
		char c;
		read( pipe_[0], &c, 1 );

		Message* msg = 0;
		ScopedLock l(mutex_);
		if (!msgQueue_.empty())
		{
			msg = msgQueue_.front();
			msgQueue_.pop();
		}
		return std::auto_ptr<Message>(msg);
	}
};

#endif /* INTERTHREADCOMM_H_ */
