/*
 * lockable.h
 *
 *  Created on: 11.08.2009
 *      Author: afestini
 *
 *  Classes derived from Lockable can be mutex locked in three ways:
 *
 *  1) manually call lock()/unlock()					to lock during longer pieces of code
 *  2) pass to a ScopedLock: ScopedLock l(instance);	to lock for a single block
 *  3) route access through safeAccess(),				to lock for one single access
 *  	e.g. object.safeAccess()->someFunction(); instead of object.someFunction();
 *  	or int x = object.safeAccess()->someInt; instead of object.someInt;
 */

#ifndef LOCKABLE_H_
#define LOCKABLE_H_

#include <pthread.h>

template<class T>
class LockedAccess
{
	T& obj;
public:
	LockedAccess(T& object) : obj(object) {obj.lock();}
	~LockedAccess() {obj.unlock();}
	T* operator->() { return &obj; }
};



template<class T>
class Lockable
{
	friend class ScopedLock;
	pthread_mutex_t mutex;

protected:
	Lockable() { pthread_mutex_init(&mutex, 0); }
	~Lockable() { pthread_mutex_destroy(&mutex); }

public:
	//Automatically lock before and unlock after access
	LockedAccess<T> safeAccess() { return LockedAccess<T>(static_cast<T*>(*this) ); }

	//Public for manual locking/unlocking
	void lock() { pthread_mutex_lock(&mutex); }
	void unlock() { pthread_mutex_unlock(&mutex); }
};


#endif /* LOCKABLE_H_ */
