/*
 * sharedptr.h
 *
 *  Created on: 27.07.2009
 *      Author: afestini
 */

#ifndef SHAREDPTR_H_
#define SHAREDPTR_H_

#include <pthread.h>

template<bool cond, class A, class B> struct SelectType { typedef A Type; };
template<class A, class B> struct SelectType<false, A, B> { typedef B Type; };

template<class Base, class Derived>
class InheritanceTest
{
	typedef char Yes;
	struct No { char x[2]; };
	static Yes test(const Base&);
	static No test(...);
	static const Derived& derived();
public:
	enum { isBase = sizeof(test(derived())) == sizeof(Yes) };
};

struct RefCounted
{
	unsigned ref;
	RefCounted() : ref(0) {}
};

class IntCounter
{
public:
	static void create() {}

	template<class SP>
	static void addRef(const SP& sp) { ++sp.get()->ref; }

	template<class SP>
	static bool subRef(const SP& sp) { return --sp.get()->ref == 0; }

	static void copy(const IntCounter&) {}
	static void destroy() {}
};

class ExtCounter
{
	unsigned *ref;
public:
	explicit ExtCounter() : ref(0) {}
	explicit ExtCounter(const ExtCounter& ec) : ref(ec.ref) {}

	void create() { ref = new unsigned(0); }
	void destroy() { delete ref; }

	template<class SP>
	void addRef(const SP&) { ++*ref; }

	template<class SP>
	bool subRef(const SP&) { return --*ref == 0; }

	void copy(const ExtCounter& p) { ref = p.ref; }
};

class SingleThreaded
{
public:
	static void create() {}
	static void destroy(){}
	static void lock() {}
	static void unlock() {}
	static void copy(const SingleThreaded&) {}
};

class MultiThreaded
{
	pthread_mutex_t *mutex;

public:
	explicit MultiThreaded() : mutex(0) {}
	explicit MultiThreaded(const MultiThreaded& mt) : mutex(mt.mutex) {}

	void create()
	{
		mutex = new pthread_mutex_t;
		pthread_mutex_init(mutex,0);
	}

	void lock()	{ pthread_mutex_lock(mutex); }
	void unlock() { pthread_mutex_unlock(mutex); }
	void copy(const MultiThreaded& sp) { mutex = sp.mutex; }
	void destroy(){ unlock(); pthread_mutex_destroy(mutex); delete mutex; }
};

template<class T, class Threading = SingleThreaded>
class SharedPtr : public SelectType< InheritanceTest<RefCounted, T>::isBase, IntCounter, ExtCounter>::Type, public Threading
{
	typedef typename SelectType< InheritanceTest<RefCounted, T>::isBase, IntCounter, ExtCounter>::Type Counter;

	T* ptr;
	class Tester { void operator delete(void*); };

	void aquire()
	{
		if (ptr)
		{
			Threading::lock();
			addRef(*this);
			Threading::unlock();
		}
	}

	void release()
	{
		if (ptr)
		{
			Threading::lock();
			if ( subRef(*this) )
			{
				Threading::destroy();
				Counter::destroy();
				delete ptr;
			}
			else
				Threading::unlock();
		}
	}

	void init()
	{
		if (ptr)
		{
			Counter::create();
			Threading::create();
			aquire();
		}
	}

	template<class U>
	SharedPtr& copy(const SharedPtr<U, Threading>& sp)
	{
		if ( ptr != sp.get() )
		{
			release();
			ptr = sp.get();
			Threading::copy(sp);
			Counter::copy(sp);
			aquire();
		}
		return *this;
	}

public:
	SharedPtr() : ptr(0) {}

	explicit SharedPtr(T* p) : ptr(p) { init(); }

	template<class U>
	explicit SharedPtr(U* p) : ptr(p) { init(); }

	SharedPtr(const SharedPtr& sp)
	: Counter(sp), Threading(sp), ptr(sp.ptr) { aquire(); }

	template<class U>
	SharedPtr(const SharedPtr<U, Threading>& sp)
	: Counter(sp), Threading(sp), ptr(sp.get()) { aquire(); }

	template<class U>
	SharedPtr(const SharedPtr<U, Threading>& sp, bool /*downcastOverload*/)
	: Counter(sp), Threading(sp), ptr(dynamic_cast<T*>(sp.get())) { aquire(); }

	~SharedPtr() { release(); }

	SharedPtr& operator=(const SharedPtr& sp) {	return copy(sp); }

	template<class U>
	SharedPtr& operator=(const SharedPtr<U, Threading>& sp) { return copy(sp); }

	//Crappy quick and dirty version
	SharedPtr& operator=(T* p) { return copy( SharedPtr(p) ); }

	template<class U>
	SharedPtr& operator=(U* p) { return copy( SharedPtr<U, Threading>(p) ); }

    bool operator==(const SharedPtr& p) const { return ptr == p.ptr; }
    bool operator!=(const SharedPtr& p) const { return ptr != p.ptr; }
    bool operator<(const SharedPtr& p) const { return ptr < p.ptr; }

	T* get() const { return ptr; }

	template<class Derived>
	SharedPtr<Derived, Threading> downcast()
	{
		return SharedPtr<Derived,Threading>(*this, true);
	}

    T* operator->() const { return ptr; }
    T& operator*() const { return *ptr; }
    operator Tester*() const { static Tester test; return ptr ? &test : 0; }
};


#endif /* SHAREDPTR_H_ */
